@@ -1,4 +1,5 @@
 class BaseConverter(object):
+    IGNORE = 'converter_ignore'
     def __init__(self):
         self.format_name = None
     def check_format(self,*args,**kwargs):
@@ -13,4 +14,6 @@ class BaseConverter(object):
         err_msg = 'Converter for %s format has no method convert_line' %\
             self.format_name
         raise NotImplementedError(err_msg)
-    
\ No newline at end of file
+    
+    def addl_operation_for_unique_variant (self, wdict, wdict_no):
+        pass
